# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/master/LICENSE

from __future__ import absolute_import

import sys

import pytest

import numpy
import awkward1


def test_same_categories():
    categories = awkward1.Array(["one", "two", "three"])
    index1 = awkward1.layout.Index64(numpy.array([0, 2, 2, 1, 2, 0, 1, 0], dtype=numpy.int64))
    index2 = awkward1.layout.Index64(numpy.array([1, 1, 2, 1, 0, 0, 0, 1], dtype=numpy.int64))
    categorical1 = awkward1.layout.IndexedArray64(index1, categories.layout, parameters={"__array__": "categorical"})
    categorical2 = awkward1.layout.IndexedArray64(index2, categories.layout, parameters={"__array__": "categorical"})
    array1 = awkward1.Array(categorical1)
    array2 = awkward1.Array(categorical2)
    assert (array1 == array2).tolist() == [False, False, True, True, False, True, False, False]


def test_different_categories():
    categories1 = awkward1.Array(["one", "two", "three"])
    categories2 = awkward1.Array(["three", "two", "one"])
    index1 = awkward1.layout.Index64(numpy.array([0, 2, 2, 1, 2, 0, 1, 0], dtype=numpy.int64))
    index2 = awkward1.layout.Index64(numpy.array([1, 1, 2, 1, 0, 0, 0, 1], dtype=numpy.int64))
    categorical1 = awkward1.layout.IndexedArray64(index1, categories1.layout, parameters={"__array__": "categorical"})
    categorical2 = awkward1.layout.IndexedArray64(index2, categories2.layout, parameters={"__array__": "categorical"})
    array1 = awkward1.Array(categorical1)
    array2 = awkward1.Array(categorical2)
    assert (array1 == array2).tolist() == [False, False, False, True, True, False, False, False]


def test_one_extra():
    categories1 = awkward1.Array(["one", "two", "three", "four"])
    categories2 = awkward1.Array(["three", "two", "one"])
    index1 = awkward1.layout.Index64(numpy.array([0, 3, 0, 1, 2, 3, 1, 0], dtype=numpy.int64))
    index2 = awkward1.layout.Index64(numpy.array([1, 1, 2, 1, 0, 0, 0, 1], dtype=numpy.int64))
    categorical1 = awkward1.layout.IndexedArray64(index1, categories1.layout, parameters={"__array__": "categorical"})
    categorical2 = awkward1.layout.IndexedArray64(index2, categories2.layout, parameters={"__array__": "categorical"})
    array1 = awkward1.Array(categorical1)
    array2 = awkward1.Array(categorical2)
    assert (array1 == array2).tolist() == [False, False, True, True, True, False, False, False]


def test_two_extra():
    categories1 = awkward1.Array(["one", "two", "three"])
    categories2 = awkward1.Array(["four", "three", "two", "one"])
    index1 = awkward1.layout.Index64(numpy.array([0, 2, 2, 1, 2, 0, 1, 0], dtype=numpy.int64))
    index2 = awkward1.layout.Index64(numpy.array([1, 3, 1, 1, 3, 3, 2, 1], dtype=numpy.int64))
    categorical1 = awkward1.layout.IndexedArray64(index1, categories1.layout, parameters={"__array__": "categorical"})
    categorical2 = awkward1.layout.IndexedArray64(index2, categories2.layout, parameters={"__array__": "categorical"})
    array1 = awkward1.Array(categorical1)
    array2 = awkward1.Array(categorical2)
    assert (array1 == array2).tolist() == [False, False, True, False, False, True, True, False]


def test_option_same_categories():
    categories = awkward1.Array(["one", "two", "three"])
    index1 = awkward1.layout.Index64(numpy.array([0, 2, 2, 1, -1,  0, 1, 0], dtype=numpy.int64))
    index2 = awkward1.layout.Index64(numpy.array([1, 1, 2, 1, -1, -1, 0, 1], dtype=numpy.int64))
    categorical1 = awkward1.layout.IndexedOptionArray64(index1, categories.layout, parameters={"__array__": "categorical"})
    categorical2 = awkward1.layout.IndexedOptionArray64(index2, categories.layout, parameters={"__array__": "categorical"})
    array1 = awkward1.Array(categorical1)
    array2 = awkward1.Array(categorical2)
    assert (array1 == array2).tolist() == [False, False, True, True, True, False, False, False]


def test_option_different_categories():
    categories1 = awkward1.Array(["one", "two", "three"])
    categories2 = awkward1.Array(["three", "two", "one"])
    index1 = awkward1.layout.Index64(numpy.array([0, 2, 2, 1, -1, -1, 1, 0], dtype=numpy.int64))
    index2 = awkward1.layout.Index64(numpy.array([1, 1, 2, 1,  0, -1, 0, 1], dtype=numpy.int64))
    categorical1 = awkward1.layout.IndexedOptionArray64(index1, categories1.layout, parameters={"__array__": "categorical"})
    categorical2 = awkward1.layout.IndexedOptionArray64(index2, categories2.layout, parameters={"__array__": "categorical"})
    array1 = awkward1.Array(categorical1)
    array2 = awkward1.Array(categorical2)
    assert (array1 == array2).tolist() == [False, False, False, True, False, True, False, False]


def test_option_one_extra():
    categories1 = awkward1.Array(["one", "two", "three", "four"])
    categories2 = awkward1.Array(["three", "two", "one"])
    index1 = awkward1.layout.Index64(numpy.array([0, -1,  0, 1, 2, 3, 1, 0], dtype=numpy.int64))
    index2 = awkward1.layout.Index64(numpy.array([1, -1, -1, 1, 0, 0, 0, 1], dtype=numpy.int64))
    categorical1 = awkward1.layout.IndexedOptionArray64(index1, categories1.layout, parameters={"__array__": "categorical"})
    categorical2 = awkward1.layout.IndexedOptionArray64(index2, categories2.layout, parameters={"__array__": "categorical"})
    array1 = awkward1.Array(categorical1)
    array2 = awkward1.Array(categorical2)
    assert (array1 == array2).tolist() == [False, True, False, True, True, False, False, False]


def test_option_two_extra():
    categories1 = awkward1.Array(["one", "two", "three"])
    categories2 = awkward1.Array(["four", "three", "two", "one"])
    index1 = awkward1.layout.Index64(numpy.array([0, -1, -1, 1, 2, 0, 1, 0], dtype=numpy.int64))
    index2 = awkward1.layout.Index64(numpy.array([1, -1,  1, 1, 3, 3, 2, 1], dtype=numpy.int64))
    categorical1 = awkward1.layout.IndexedOptionArray64(index1, categories1.layout, parameters={"__array__": "categorical"})
    categorical2 = awkward1.layout.IndexedOptionArray64(index2, categories2.layout, parameters={"__array__": "categorical"})
    array1 = awkward1.Array(categorical1)
    array2 = awkward1.Array(categorical2)
    assert (array1 == array2).tolist() == [False, True, False, False, False, True, True, False]


def test_to_categorical_numbers():
    array = awkward1.Array([1.1, 2.2, 3.3, 1.1, 2.2, 3.3, 1.1, 2.2, 3.3])
    assert not awkward1.is_categorical(array)
    categorical = awkward1.to_categorical(array)
    assert awkward1.is_categorical(categorical)
    assert awkward1.to_list(array) == categorical.tolist()
    assert awkward1.to_list(categorical.layout.content) == [1.1, 2.2, 3.3]
    not_categorical = awkward1.from_categorical(categorical)
    assert not awkward1.is_categorical(not_categorical)
    assert awkward1.categories(categorical).tolist() == [1.1, 2.2, 3.3]


def test_to_categorical_nested():
    array = awkward1.Array([["one", "two", "three"], [], ["one", "two"], ["three"]])
    assert not awkward1.is_categorical(array)
    categorical = awkward1.to_categorical(array)
    assert awkward1.is_categorical(categorical)
    assert awkward1.to_list(array) == categorical.tolist()
    not_categorical = awkward1.from_categorical(categorical)
    assert not awkward1.is_categorical(not_categorical)
    assert awkward1.categories(categorical).tolist() == ["one", "two", "three"]


def test_to_categorical():
    array = awkward1.Array(["one", "two", "three", "one", "two", "three", "one", "two", "three"])
    assert not awkward1.is_categorical(array)
    categorical = awkward1.to_categorical(array)
    assert awkward1.is_categorical(categorical)
    assert awkward1.to_list(array) == categorical.tolist()
    assert awkward1.to_list(categorical.layout.content) == ["one", "two", "three"]
    not_categorical = awkward1.from_categorical(categorical)
    assert not awkward1.is_categorical(not_categorical)
    assert awkward1.categories(categorical).tolist() == ["one", "two", "three"]


def test_to_categorical_none():
    array = awkward1.Array(["one", "two", "three", None, "one", "two", "three", None, "one", "two", "three", None])
    assert not awkward1.is_categorical(array)
    categorical = awkward1.to_categorical(array)
    assert awkward1.is_categorical(categorical)
    assert awkward1.to_list(array) == categorical.tolist()
    assert awkward1.to_list(categorical.layout.content) == ["one", "two", "three"]
    not_categorical = awkward1.from_categorical(categorical)
    assert not awkward1.is_categorical(not_categorical)
    assert awkward1.categories(categorical).tolist() == ["one", "two", "three"]


def test_to_categorical_masked():
    content = awkward1.Array(["one", "two", "three", "one", "one", "two", "three", "two", "one", "two", "three", "three"]).layout
    mask = awkward1.layout.Index8(numpy.array([False, False, False, True, False, False, False, True, False, False, False, True]))
    array = awkward1.Array(awkward1.layout.ByteMaskedArray(mask, content, valid_when=False))
    assert not awkward1.is_categorical(array)
    categorical = awkward1.to_categorical(array)
    assert awkward1.is_categorical(categorical)
    assert awkward1.to_list(array) == categorical.tolist()
    assert awkward1.to_list(categorical.layout.content) == ["one", "two", "three"]
    not_categorical = awkward1.from_categorical(categorical)
    assert not awkward1.is_categorical(not_categorical)
    assert awkward1.categories(categorical).tolist() == ["one", "two", "three"]


def test_to_categorical_masked():
    content = awkward1.Array(["one", "two", "three", "one", "one", "two", "three", "two"]).layout
    index = awkward1.layout.Index64(numpy.array([0, 1, 2, 3, 4, 5, 6, 7, 0, 1, 2, 3], dtype=numpy.int64))
    indexedarray = awkward1.layout.IndexedArray64(index, content)
    mask = awkward1.layout.Index8(numpy.array([False, False, False, True, False, False, False, True, False, False, False, True]))
    array = awkward1.Array(awkward1.layout.ByteMaskedArray(mask, indexedarray, valid_when=False))
    assert not awkward1.is_categorical(array)
    categorical = awkward1.to_categorical(array)
    assert awkward1.is_categorical(categorical)
    assert awkward1.to_list(array) == categorical.tolist()
    assert awkward1.to_list(categorical.layout.content) == ["one", "two", "three"]
    not_categorical = awkward1.from_categorical(categorical)
    assert not awkward1.is_categorical(not_categorical)
    assert awkward1.categories(categorical).tolist() == ["one", "two", "three"]


def test_typestr():
    if not awkward1._util.py27:
        assert str(awkward1.type(awkward1.to_categorical(awkward1.Array([1.1, 2.2, 2.2, 3.3])))) == "4 * categorical[type=float64]"
        assert str(awkward1.type(awkward1.to_categorical(awkward1.Array([1.1, 2.2, None, 2.2, 3.3])))) == "5 * categorical[type=?float64]"
        assert str(awkward1.type(awkward1.to_categorical(awkward1.Array(["one", "two", "two", "three"])))) == "4 * categorical[type=string]"
        assert str(awkward1.type(awkward1.to_categorical(awkward1.Array(["one", "two", None, "two", "three"])))) == "5 * categorical[type=option[string]]"


def test_zip():
    x = awkward1.Array([1.1, 2.2, 3.3])
    y = awkward1.Array(["one", "two", "three"])
    assert awkward1.zip({"x": x, "y": y}).tolist() == [{"x": 1.1, "y": "one"}, {"x": 2.2, "y": "two"}, {"x": 3.3, "y": "three"}]
    y = awkward1.to_categorical(y)
    assert awkward1.zip({"x": x, "y": y}).tolist() == [{"x": 1.1, "y": "one"}, {"x": 2.2, "y": "two"}, {"x": 3.3, "y": "three"}]


pyarrow = pytest.importorskip("pyarrow")


def test_arrow_nomask():
    array = awkward1.Array([1.1, 2.2, 3.3, 4.4, None])
    assert str(awkward1.type(awkward1.from_arrow(awkward1.to_arrow(array)))) == "5 * ?float64"
    assert str(awkward1.type(awkward1.from_arrow(awkward1.to_arrow(array[:-1])))) == "4 * ?float64"



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/master/LICENSE

from __future__ import absolute_import

import sys
import itertools

import pytest
import numpy

import awkward1

def test():
    assert awkward1.to_list(awkward1.prod(awkward1.Array([[[2, 3, 5]], [[7], [11]], [[]]]), axis=-1)) == [[30], [7, 11], [1]]

    assert awkward1.to_list(awkward1.prod(awkward1.Array([[[2, 3, 5]], [[7], [11]], []]), axis=-1)) == [[30], [7, 11], []]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/master/LICENSE

from __future__ import absolute_import

import distutils.version
import warnings

import awkward1.layout
import awkward1._util
import awkward1.operations.convert
import awkward1.operations.structure
import awkward1.nplike


np = awkward1.nplike.NumpyMetadata.instance()


def register():
    global AwkwardDtype
    try:
        pandas = get_pandas()
    except ImportError:
        pass
    else:
        if AwkwardDtype is None:
            get_dtype()
        if issubclass(PandasMixin, PandasNotImportedYet):
            PandasMixin.__bases__ = (pandas.api.extensions.ExtensionArray,)


def vote():
    global AwkwardDtype

    if AwkwardDtype is None:
        raise RuntimeError(
            "Use of Awkward Arrays in Pandas Series and DataFrames is deprecated."
            "\nFor now, you can\n\n    ak.pandas.register()\n\n"
            "to get this feature, but it will be removed in Awkward1 0.3.0. "
            "(see https://github.com/scikit-hep/awkward-1.0/issues/350)"
        )


checked_version = False


def get_pandas():
    global checked_version
    try:
        import pandas
    except ImportError:
        raise ImportError(
            """install the 'pandas' package with:

    pip install pandas --upgrade

or

    conda install pandas"""
        )
    else:
        if not checked_version and distutils.version.LooseVersion(
            pandas.__version__
        ) < distutils.version.LooseVersion("0.24"):
            raise ImportError(
                "awkward1 can only work with pandas 0.24 or later "
                "(you have version {0})".format(pandas.__version__)
            )
        checked_version = True
        return pandas


AwkwardDtype = None


def get_dtype():
    import awkward1.highlevel

    pandas = get_pandas()

    global AwkwardDtype
    if AwkwardDtype is None:

        @pandas.api.extensions.register_extension_dtype
        class AwkwardDtype(pandas.api.extensions.ExtensionDtype):
            name = "awkward1"
            type = awkward1.highlevel.Array
            kind = "O"
            base = np.dtype("O")

            @classmethod
            def construct_from_string(cls, string):
                if string == cls.name:
                    return cls()
                else:
                    raise TypeError(
                        "cannot construct a {0} from {1}".format(cls, string)
                        + awkward1._util.exception_suffix(__file__)
                    )

            @classmethod
            def construct_array_type(cls):
                return awkward1.highlevel.Array

    return AwkwardDtype


class PandasNotImportedYet(object):
    pass


class NoFields(object):
    def __str__(self):
        return "(no fields)"

    def __eq__(self, other):
        return other is NoFields or isinstance(other, NoFields)

    def __hash__(self):
        return hash(NoFields)


class PandasMixin(PandasNotImportedYet):
    @property
    def _typ(self):
        vote()
        return "dataframe"

    @property
    def columns(self):
        if self.layout.numfields >= 0:
            return self.layout.keys()
        else:
            return [NoFields()]

    def _ixs(self, i, axis):
        vote()
        if self.layout.numfields >= 0:
            return get_pandas().Series(self[str(i)])
        else:
            return get_pandas().Series(self)

    # REQUIRED by Pandas:

    @classmethod
    def _from_sequence(cls, scalars, *args, **kwargs):
        # https://pandas.pydata.org/pandas-docs/version/1.0.0/reference/api/pandas.api.extensions.ExtensionArray._from_sequence.html
        vote()
        dtype, copy = awkward1._util.extra(
            args, kwargs, [("dtype", None), ("copy", False)]
        )
        return awkward1.operations.convert.from_iter(scalars)

    @classmethod
    def _from_factorized(cls, values, original):
        # https://pandas.pydata.org/pandas-docs/version/1.0.0/reference/api/pandas.api.extensions.ExtensionArray._from_factorized.html
        vote()
        raise NotImplementedError(
            "_from_factorized"
            + awkward1._util.exception_suffix(__file__)
        )

    # __getitem__(self)
    # __len__(self)

    @property
    def dtype(self):
        # https://pandas.pydata.org/pandas-docs/version/1.0.0/reference/api/pandas.api.extensions.ExtensionArray.dtype.html
        if awkward1._util.called_by_module(
            "pandas"
        ) and not awkward1._util.called_by_module("dask"):
            vote()
            if isinstance(self.layout, awkward1.partition.PartitionedArray):
                raise ValueError(
                    "partitioned arrays cannot be Pandas columns; "
                    "try ak.repartition(array, None)"
                    + awkward1._util.exception_suffix(__file__)
                )
            else:
                return AwkwardDtype()

        else:
            return np.dtype("O")

    @property
    def nbytes(self):
        # https://pandas.pydata.org/pandas-docs/version/1.0.0/reference/api/pandas.api.extensions.ExtensionArray.nbytes.html
        return self._layout.nbytes

    @property
    def ndim(self):
        # https://pandas.pydata.org/pandas-docs/version/1.0.0/reference/api/pandas.api.extensions.ExtensionArray.nbytes.html
        return 1

    @property
    def shape(self):
        # https://pandas.pydata.org/pandas-docs/version/1.0.0/reference/api/pandas.api.extensions.ExtensionArray.shape.html
        return (len(self),)

    def isna(self):
        # https://pandas.pydata.org/pandas-docs/version/1.0.0/reference/api/pandas.api.extensions.ExtensionArray.isna.html
        vote()
        nplike = awkward1.nplike.of(self)
        return nplike.array(awkward1.operations.structure.is_none(self))

    def take(self, indices, *args, **kwargs):
        # https://pandas.pydata.org/pandas-docs/version/1.0.0/reference/api/pandas.api.extensions.ExtensionArray.take.html
        allow_fill, fill_value = awkward1._util.extra(
            args, kwargs, [("allow_fill", False), ("fill_value", None)]
        )
        vote()

        nplike = awkward1.nplike.of(self)
        if allow_fill:
            content1 = self.layout
            if isinstance(content1, awkward1.partition.PartitionedArray):
                content1 = content1.toContent()

            indices = nplike.asarray(indices, dtype=np.int64)
            if fill_value is None:
                index = awkward1.layout.Index64(indices)
                layout = awkward1.layout.IndexedOptionArray64(
                    index, content1, parameters=self.layout.parameters
                )
                return awkward1._util.wrap(layout, awkward1._util.behaviorof(self))

            else:
                tags = (indices >= 0).view(np.int8)
                index = indices.copy()
                index[~tags] = 0
                content0 = awkward1.operations.convert.from_iter(
                    [fill_value], highlevel=False
                )
                tags = awkward1.layout.Index8(tags)
                index = awkward1.layout.Index64(index)
                layout = awkward1.layout.UnionArray8_64(
                    tags, index, [content0, content1]
                )
                return awkward1._util.wrap(layout, awkward1._util.behaviorof(self))

        else:
            return self[indices]

    def copy(self):
        # https://pandas.pydata.org/pandas-docs/version/1.0.0/reference/api/pandas.api.extensions.ExtensionArray.copy.html
        return awkward1._util.wrap(
            self._layout.deep_copy(
                copyarrays=True, copyindexes=True, copyidentities=True
            ),
            awkward1._util.behaviorof(self),
        )

    @classmethod
    def _concat_same_type(cls, to_concat):
        # https://pandas.pydata.org/pandas-docs/version/1.0.0/reference/api/pandas.api.extensions.ExtensionArray._concat_same_type.html
        vote()
        return awkward1.operations.structure.concatenate(to_concat)

    # RECOMMENDED for performance:

    # def fillna(self, *args, **kwargs):
    #     # https://pandas.pydata.org/pandas-docs/version/1.0.0/reference/api/pandas.api.extensions.ExtensionArray.fillna.html
    #     value, method, limit = awkward1._util.extra(args, kwargs, [
    #         ("value", None),
    #         ("method", None),
    #         ("limit", None)])
    #     vote()
    #     raise NotImplementedError
    #
    # def dropna(self):
    #     # https://pandas.pydata.org/pandas-docs/version/1.0.0/reference/api/pandas.api.extensions.ExtensionArray.dropna.html
    #     vote()
    #     raise NotImplementedError
    #
    # def unique(self):
    #     # https://pandas.pydata.org/pandas-docs/version/1.0.0/reference/api/pandas.api.extensions.ExtensionArray.unique.html
    #     vote()
    #     raise NotImplementedError
    #
    # def factorize(self, na_sentinel):
    #     # https://pandas.pydata.org/pandas-docs/version/1.0.0/reference/api/pandas.api.extensions.ExtensionArray.factorize.html
    #     vote()
    #     raise NotImplementedError
    #
    # def _values_for_factorize(self):
    #     # https://pandas.pydata.org/pandas-docs/version/1.0.0/reference/api/pandas.api.extensions.ExtensionArray._values_for_factorize.html
    #     vote()
    #     raise NotImplementedError
    #
    # def argsort(self, *args, **kwargs):
    #     # https://pandas.pydata.org/pandas-docs/version/1.0.0/reference/api/pandas.api.extensions.ExtensionArray.argsort.html
    #     ascending, kind = awkward1._util.extra(args, kwargs, [
    #         ("ascending", True),
    #         ("kind", "quicksort")])   # "quicksort", "mergesort", "heapsort"
    #     vote()
    #     raise NotImplementedError
    #
    # def _values_for_argsort(self):
    #     # https://pandas.pydata.org/pandas-docs/version/1.0.0/reference/api/pandas.api.extensions.ExtensionArray._values_for_argsort.html
    #     vote()
    #     raise NotImplementedError
    #
    # def searchsorted(self, value, *args, **kwargs):
    #     # https://pandas.pydata.org/pandas-docs/version/1.0.0/reference/api/pandas.api.extensions.ExtensionArray.searchsorted.html
    #     side, sorter = awkward1._util.extra(args, kwargs, [
    #         ("side", "left"),
    #         ("sorter", None)])
    #     vote()
    #     raise NotImplementedError
    #
    # def _reduce(self, name, *args, **kwargs):
    #     # https://pandas.pydata.org/pandas-docs/version/1.0.0/reference/api/pandas.api.extensions.ExtensionArray._reduce.html
    #     skipna, = awkward1._util.extra(args, kwargs, [
    #         ("skipna", True)])
    #     vote()
    #     raise NotImplementedError


def df(array, how="inner", levelname=lambda i: "sub" * i + "entry", anonymous="values"):
    warnings.warn(
        "ak.pandas.df is deprecated, will be removed in 0.3.0. Use\n\n"
        "    ak.to_pandas(array)\n\ninstead.",
        DeprecationWarning,
    )
    return awkward1.operations.convert.to_pandas(
        array, how=how, levelname=levelname, anonymous=anonymous
    )

def dfs(array, levelname=lambda i: "sub" * i + "entry", anonymous="values"):
    warnings.warn(
        "ak.pandas.dfs is deprecated, will be removed in 0.3.0. Use\n\n"
        "    ak.to_pandas(array, how=None)\n\ninstead.",
        DeprecationWarning,
    )
    return awkward1.operations.convert.to_pandas(
        array, how=None, levelname=levelname, anonymous=anonymous
    )



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/master/LICENSE

from __future__ import absolute_import

import sys
import itertools
import collections

import pytest
import numpy

import awkward1
import awkward1.behaviors.string

py27 = (sys.version_info[0] < 3)

def test_fromnumpy():
    a = numpy.arange(2*3*5).reshape((2, 3, 5))
    b = awkward1.from_numpy(a)
    assert awkward1.to_list(a) == awkward1.to_list(b)

def test_highlevel():
    a = awkward1.Array([[1.1, 2.2, 3.3], [], [4.4, 5.5], [6.6], [7.7, 8.8, 9.9]], check_valid=True)
    assert repr(a) == "<Array [[1.1, 2.2, 3.3], ... [7.7, 8.8, 9.9]] type='5 * var * float64'>"
    assert str(a) == "[[1.1, 2.2, 3.3], [], [4.4, 5.5], [6.6], [7.7, 8.8, 9.9]]"

    b = awkward1.Array(numpy.arange(100, dtype=numpy.int32), check_valid=True)
    assert repr(b) == "<Array [0, 1, 2, 3, 4, ... 95, 96, 97, 98, 99] type='100 * int32'>"
    assert str(b) == "[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, ... 89, 90, 91, 92, 93, 94, 95, 96, 97, 98, 99]"

    c = awkward1.Array('[{"one": 3.14, "two": [1.1, 2.2]}, {"one": 99.9, "two": [-3.1415926]}]', check_valid=True)
    assert repr(c) == "<Array [{one: 3.14, two: [1.1, ... -3.14]}] type='2 * {\"one\": float64, \"two\": va...'>"
    assert str(c) == "[{one: 3.14, two: [1.1, 2.2]}, {one: 99.9, two: [-3.14]}]"

class Dummy(awkward1.highlevel.Array):
    pass

def test_string1():
    a = awkward1.Array(numpy.array([ord(x) for x in "hey there"], dtype=numpy.uint8), check_valid=True)
    a.__class__ = awkward1.behaviors.string.ByteBehavior
    assert str(a) == str(b"hey there")
    assert repr(a) == repr(b"hey there")

def test_string2():
    content = awkward1.layout.NumpyArray(numpy.array([ord(x) for x in "heythere"], dtype=numpy.uint8))
    listoffsetarray = awkward1.layout.ListOffsetArray64(awkward1.layout.Index64(numpy.array([0, 3, 3, 8])), content)
    a = awkward1.Array(listoffsetarray, check_valid=True)

    assert isinstance(a, awkward1.Array)
    assert not isinstance(a, awkward1.behaviors.string.StringBehavior)
    assert awkward1.to_list(a) == [[104, 101, 121], [], [116, 104, 101, 114, 101]]

    assert repr(awkward1.type(a)) == "3 * var * uint8"
    assert repr(awkward1.type(a[0])) == "3 * uint8"
    assert repr(awkward1.type(a[1])) == "0 * uint8"
    assert repr(awkward1.type(a[2])) == "5 * uint8"

    assert repr(a) == "<Array [[104, 101, 121], ... 101, 114, 101]] type='3 * var * uint8'>"
    assert str(a) == "[[104, 101, 121], [], [116, 104, 101, 114, 101]]"
    assert repr(a[0]) == "<Array [104, 101, 121] type='3 * uint8'>"
    assert repr(a[1]) == "<Array [] type='0 * uint8'>"
    assert repr(a[2]) == "<Array [116, 104, 101, 114, 101] type='5 * uint8'>"

    content = awkward1.layout.NumpyArray(numpy.array([ord(x) for x in "heythere"], dtype=numpy.uint8), parameters={"__array__": "char", "encoding": "utf-8"})
    listoffsetarray = awkward1.layout.ListOffsetArray64(awkward1.layout.Index64(numpy.array([0, 3, 3, 8])), content, parameters={"__array__": "string"})
    a = awkward1.Array(listoffsetarray, check_valid=True)

    a = awkward1.Array(listoffsetarray, check_valid=True)
    assert isinstance(a, awkward1.Array)
    assert awkward1.to_list(a) == ['hey', '', 'there']

    if py27:
        assert str(a) == "[u'hey', u'', u'there']"
        assert repr(a[0]) == "u'hey'"
        assert repr(a[1]) == "u''"
        assert repr(a[2]) == "u'there'"
    else:
        assert str(a) == "['hey', '', 'there']"
        assert repr(a[0]) == "'hey'"
        assert repr(a[1]) == "''"
        assert repr(a[2]) == "'there'"



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/master/LICENSE

from __future__ import absolute_import

import sys

import pytest
import numpy

import awkward1

def test():
    array = awkward1.Array([{"x": "one"}, {"x": "two"}, {"x": "three"}], check_valid=True)
    assert awkward1.to_list(array) == [{"x": "one"}, {"x": "two"}, {"x": "three"}]
    assert awkward1.to_list(awkward1.from_iter(awkward1.to_list(array))) == [{"x": "one"}, {"x": "two"}, {"x": "three"}]
    assert awkward1.to_list(array.layout) == [{"x": "one"}, {"x": "two"}, {"x": "three"}]
    assert awkward1.to_list(awkward1.from_iter(awkward1.to_list(array.layout))) == [{"x": "one"}, {"x": "two"}, {"x": "three"}]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/master/LICENSE

from __future__ import absolute_import

import sys

import pytest
import numpy

import awkward1


@pytest.mark.skipif(awkward1._util.py27 and awkward1._util.win, reason="Windows Python 2.7 inserts 'L' after numbers in repr")
def test():
    assert str(awkward1.Array([
        {"a": 1, "a b": 2}, {"a": 1, "a b": 2}, {"a": 1, "a b": 2}, {"a": 1, "a b": 2}
    ])) in (
        "[{a: 1, 'a b': 2}, {a: 1, 'a b': 2}, {a: 1, 'a b': 2}, {a: 1, 'a b': 2}]",
        "[{'a b': 2, a: 1}, {'a b': 2, a: 1}, {'a b': 2, a: 1}, {'a b': 2, a: 1}]"
    )
    assert repr(awkward1.Array([
        {"a": 1, "a b": 2}, {"a": 1, "a b": 2}, {"a": 1, "a b": 2}
    ])) in (
        "<Array [{a: 1, 'a b': 2}, ... {a: 1, 'a b': 2}] type='3 * {\"a\": int64, \"a b\": in...'>"
        "<Array [{'a b': 2, a: 1}, ... {'a b': 2, a: 1}] type='3 * {\"a b\": int64, \"a\": in...'>"
    )



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/master/LICENSE

from __future__ import absolute_import

import sys

import pytest
import numpy

import awkward1

def test_broadcast_single_bool():
    base = awkward1.Array([[{"x" : 0.1, "y" : 0.2, "z" : 0.3}, {"x" : 0.4, "y" : 0.5, "z" : 0.6}]])
    base_new1 = awkward1.with_field(base, True, "always_true")
    assert awkward1.to_list(base_new1.always_true) == [[True, True]]
    base_new2 = awkward1.with_field(base_new1, base.x > 0.3, "sometimes_true")
    assert awkward1.to_list(base_new2.always_true) == [[True, True]]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/master/LICENSE

from __future__ import absolute_import

import json

# don't import awkward1._connect._uproot in awkward1/__init__.py!
import uproot4

import awkward1._io


def can_optimize(interpretation, form):
    if isinstance(interpretation, uproot4.interpretation.objects.AsObjects):
        jsonform = json.loads(form.tojson(verbose=True))
        if (
            jsonform["class"] == "ListOffsetArray64" and
            jsonform["parameters"].get("uproot") == {
                "as": "array", "header": True, "speedbump": False
            } and
            jsonform["content"]["class"] == "ListOffsetArray64" and
            jsonform["content"]["parameters"].get("uproot") == {
                "as": "vector", "header": False
            } and
            jsonform["content"]["content"]["class"] == "NumpyArray" and
            jsonform["content"]["content"]["inner_shape"] == [] and
            (
                jsonform["content"]["content"].get("primitive") == "float64" or
                jsonform["content"]["content"].get("primitive") == "int32"
            )
        ):
            return True

    return False


def basket_array(form, data, byte_offsets, extra):
    # FIXME: uproot4_issue_90 is just a placeholder, to show how it would be done

    return awkward1._io.uproot_issue_90(
        form,
        awkward1.layout.NumpyArray(data),
        awkward1.layout.Index32(byte_offsets),
    )



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/master/LICENSE

from __future__ import absolute_import

import sys

import numpy

import awkward1.layout
import awkward1.operations.convert
import awkward1._util


def convert_to_array(layout, args, kwargs):
    out = awkward1.operations.convert.to_numpy(layout, allow_missing=False)
    if args == () and kwargs == {}:
        return out
    else:
        return numpy.array(out, *args, **kwargs)


implemented = {}


def array_function(func, types, args, kwargs):
    function = implemented.get(func)
    if function is None:
        return NotImplemented
    else:
        return function(*args, **kwargs)


def implements(numpy_function):
    def decorator(function):
        implemented[getattr(numpy, numpy_function)] = function
        return function
    return decorator


def array_ufunc(ufunc, method, inputs, kwargs):
    import awkward1.highlevel

    if method != "__call__" or len(inputs) == 0 or "out" in kwargs:
        return NotImplemented

    behavior = awkward1._util.behaviorof(*inputs)
    inputs = [
        awkward1.operations.convert.to_layout(x, allow_record=True, allow_other=True)
        for x in inputs
    ]

    def adjust(custom, inputs, kwargs):
        args = [
            awkward1._util.wrap(x, behavior)
            if isinstance(x, (awkward1.layout.Content, awkward1.layout.Record))
            else x
            for x in inputs
        ]
        out = custom(*args, **kwargs)
        if not isinstance(out, tuple):
            out = (out,)

        return tuple(
            x.layout
            if isinstance(x, (awkward1.highlevel.Array, awkward1.highlevel.Record))
            else x
            for x in out
        )

    def getfunction(inputs, depth):
        signature = (ufunc,) + tuple(
            x.parameters.get("__record__")
            if isinstance(x, awkward1.layout.Content)
            else type(x)
            for x in inputs
        )
        custom = awkward1._util.overload(behavior, signature)
        if custom is not None:
            return lambda: adjust(custom, inputs, kwargs)

        signature = (ufunc,) + tuple(
            x.parameters.get("__array__")
            if isinstance(x, awkward1.layout.Content)
            else type(x)
            for x in inputs
        )
        custom = awkward1._util.overload(behavior, signature)
        if custom is not None:
            return lambda: adjust(custom, inputs, kwargs)

        if all(
            isinstance(x, awkward1.layout.NumpyArray)
            or not isinstance(
                x, (awkward1.layout.Content, awkward1.partition.PartitionedArray)
            )
            for x in inputs
        ):
            nplike = awkward1.nplike.of(*inputs)
            result = getattr(ufunc, method)(
                *[nplike.asarray(x) for x in inputs], **kwargs
            )
            return lambda: (awkward1.layout.NumpyArray(result),)

        return None

    out = awkward1._util.broadcast_and_apply(inputs, getfunction, behavior)
    assert isinstance(out, tuple) and len(out) == 1
    return awkward1._util.wrap(out[0], behavior)


try:
    NDArrayOperatorsMixin = numpy.lib.mixins.NDArrayOperatorsMixin

except AttributeError:
    from numpy.core import umath as um

    def _disables_array_ufunc(obj):
        try:
            return obj.__array_ufunc__ is None
        except AttributeError:
            return False

    def _binary_method(ufunc, name):
        def func(self, other):
            if _disables_array_ufunc(other):
                return NotImplemented
            return ufunc(self, other)

        func.__name__ = "__{}__".format(name)
        return func

    def _reflected_binary_method(ufunc, name):
        def func(self, other):
            if _disables_array_ufunc(other):
                return NotImplemented
            return ufunc(other, self)

        func.__name__ = "__r{}__".format(name)
        return func

    def _inplace_binary_method(ufunc, name):
        def func(self, other):
            return ufunc(self, other, out=(self,))

        func.__name__ = "__i{}__".format(name)
        return func

    def _numeric_methods(ufunc, name):
        return (
            _binary_method(ufunc, name),
            _reflected_binary_method(ufunc, name),
            _inplace_binary_method(ufunc, name),
        )

    def _unary_method(ufunc, name):
        def func(self):
            return ufunc(self)

        func.__name__ = "__{}__".format(name)
        return func

    class NDArrayOperatorsMixin(object):
        __lt__ = _binary_method(um.less, "lt")
        __le__ = _binary_method(um.less_equal, "le")
        __eq__ = _binary_method(um.equal, "eq")
        __ne__ = _binary_method(um.not_equal, "ne")
        __gt__ = _binary_method(um.greater, "gt")
        __ge__ = _binary_method(um.greater_equal, "ge")

        __add__, __radd__, __iadd__ = _numeric_methods(um.add, "add")
        __sub__, __rsub__, __isub__ = _numeric_methods(um.subtract, "sub")
        __mul__, __rmul__, __imul__ = _numeric_methods(um.multiply, "mul")
        if sys.version_info.major < 3:
            __div__, __rdiv__, __idiv__ = _numeric_methods(um.divide, "div")
        __truediv__, __rtruediv__, __itruediv__ = _numeric_methods(
            um.true_divide, "truediv"
        )
        __floordiv__, __rfloordiv__, __ifloordiv__ = _numeric_methods(
            um.floor_divide, "floordiv"
        )
        __mod__, __rmod__, __imod__ = _numeric_methods(um.remainder, "mod")
        if hasattr(um, "divmod"):
            __divmod__ = _binary_method(um.divmod, "divmod")
            __rdivmod__ = _reflected_binary_method(um.divmod, "divmod")
        __pow__, __rpow__, __ipow__ = _numeric_methods(um.power, "pow")
        __lshift__, __rlshift__, __ilshift__ = _numeric_methods(
            um.left_shift, "lshift"
        )
        __rshift__, __rrshift__, __irshift__ = _numeric_methods(
            um.right_shift, "rshift"
        )
        __and__, __rand__, __iand__ = _numeric_methods(um.bitwise_and, "and")
        __xor__, __rxor__, __ixor__ = _numeric_methods(um.bitwise_xor, "xor")
        __or__, __ror__, __ior__ = _numeric_methods(um.bitwise_or, "or")

        __neg__ = _unary_method(um.negative, "neg")
        if hasattr(um, "positive"):
            __pos__ = _unary_method(um.positive, "pos")
            __abs__ = _unary_method(um.absolute, "abs")
            __invert__ = _unary_method(um.invert, "invert")



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/master/LICENSE

import sys

import pytest
import numpy

import awkward1

def test_fillna_empty_array():
    empty = awkward1.layout.EmptyArray()

    assert awkward1.to_list(empty) == []
    array = empty.rpad(5, 0)
    assert awkward1.to_list(array) == [None, None, None, None, None]
    value = awkward1.layout.NumpyArray(numpy.array([10]))
    assert awkward1.to_list(array.fillna(value)) == [10, 10, 10, 10, 10]

def test_fillna_numpy_array():
    content = awkward1.layout.NumpyArray(numpy.array([[1.1, 2.2, 3.3], [4.4, 5.5, 6.6]]))
    array = content.rpad(3,0)
    assert awkward1.to_list(array) == [[1.1, 2.2, 3.3], [4.4, 5.5, 6.6], None]
    value = awkward1.layout.NumpyArray(numpy.array([0]))
    assert awkward1.to_list(array.fillna(value)) == [[1.1, 2.2, 3.3], [4.4, 5.5, 6.6], 0]

    array = content.rpad(5,1)
    assert awkward1.to_list(array) == [[1.1, 2.2, 3.3, None, None], [4.4, 5.5, 6.6, None, None]]
    value = awkward1.layout.NumpyArray(numpy.array([0]))
    assert awkward1.to_list(array.fillna(value)) == [[1.1, 2.2, 3.3, 0, 0], [4.4, 5.5, 6.6, 0, 0]]

# def test_fillna_numpy_O_array():
#     pyobject_array = awkward1.layout.NumpyArray(numpy.array([1.1, 2.2, 3.3, None, 5.5, 6.6, None]))
#     assert awkward1.to_list(optarray.fillna(0)) == [1.1, 2.2, 3.3, 0, 5.5, 6.6, 0]

def test_fillna_regular_array():
    content = awkward1.layout.NumpyArray(numpy.array([2.1, 8.4, 7.4, 1.6, 2.2, 3.4, 6.2, 5.4, 1.5, 3.9, 3.8, 3.0, 8.5, 6.9, 4.3, 3.6, 6.7, 1.8, 3.2]))
    index = awkward1.layout.Index64(numpy.array([13, 9, 13, 4, 8, 3, 15, -1, 16, 2, 8], dtype=numpy.int64))
    indexedarray = awkward1.layout.IndexedOptionArray64(index, content)
    regarray = awkward1.layout.RegularArray(indexedarray, 3)

    assert awkward1.to_list(regarray) == [[6.9, 3.9, 6.9], [2.2, 1.5, 1.6], [3.6, None, 6.7]]
    value = awkward1.layout.NumpyArray(numpy.array([666]))
    assert awkward1.to_list(regarray.fillna(value)) ==  [[6.9, 3.9, 6.9], [2.2, 1.5, 1.6], [3.6, 666, 6.7]]

def test_fillna_listarray_array():
    content = awkward1.layout.NumpyArray(numpy.array([0.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9]))
    starts  = awkward1.layout.Index64(numpy.array([0, 3, 4, 5, 8]))
    stops   = awkward1.layout.Index64(numpy.array([3, 3, 6, 8, 9]))
    listarray   = awkward1.layout.ListArray64(starts, stops, content)

    assert awkward1.to_list(listarray) == [[0.0, 1.1, 2.2], [], [4.4, 5.5], [5.5, 6.6, 7.7], [8.8]]
    value = awkward1.layout.NumpyArray(numpy.array([55]))
    assert awkward1.to_list(listarray.fillna(value)) == [[0.0, 1.1, 2.2], [], [4.4, 5.5], [5.5, 6.6, 7.7], [8.8]]

def test_fillna_unionarray():
    content1 = awkward1.from_iter([[], [1.1], [2.2, 2.2]], highlevel=False)
    content2 = awkward1.from_iter([[2, 2], [1], []], highlevel=False)
    tags = awkward1.layout.Index8(numpy.array([0, 1, 0, 1, 0, 1], dtype=numpy.int8))
    index = awkward1.layout.Index64(numpy.array([0, 0, 1, 1, 2, 2], dtype=numpy.int64))
    array = awkward1.layout.UnionArray8_64(tags, index, [content1, content2])
    assert awkward1.to_list(array) == [[], [2, 2], [1.1], [1], [2.2, 2.2], []]

    padded_array = array.rpad(2, 1)
    assert awkward1.to_list(padded_array) == [[None, None], [2, 2], [1.1, None], [1, None], [2.2, 2.2], [None, None]]
    value = awkward1.layout.NumpyArray(numpy.array([777]))
    assert awkward1.to_list(padded_array.fillna(value)) == [[777, 777], [2, 2], [1.1, 777], [1, 777], [2.2, 2.2], [777, 777]]

def test_highlevel():
    array = awkward1.Array([[1.1, 2.2, None, 3.3], [], [4.4, None, 5.5]])
    assert awkward1.to_list(awkward1.fill_none(array, 999)) == [[1.1, 2.2, 999, 3.3], [], [4.4, 999, 5.5]]
    assert awkward1.to_list(awkward1.fill_none(array, [1, 2, 3])) == [[1.1, 2.2, [1, 2, 3], 3.3], [], [4.4, [1, 2, 3], 5.5]]
    assert awkward1.to_list(awkward1.fill_none(array, [])) == [[1.1, 2.2, [], 3.3], [], [4.4, [], 5.5]]
    assert awkward1.to_list(awkward1.fill_none(array, {"x": 999})) == [[1.1, 2.2, {"x": 999}, 3.3], [], [4.4, {"x": 999}, 5.5]]

    array = awkward1.Array([[1.1, 2.2, 3.3], None, [], None, [4.4, 5.5]])
    assert awkward1.to_list(awkward1.fill_none(array, 999)) == [[1.1, 2.2, 3.3], 999, [], 999, [4.4, 5.5]]
    assert awkward1.to_list(awkward1.fill_none(array, [1, 2, 3])) == [[1.1, 2.2, 3.3], [1, 2, 3], [], [1, 2, 3], [4.4, 5.5]]
    assert awkward1.to_list(awkward1.fill_none(array, {"x": 999})) == [[1.1, 2.2, 3.3], {"x": 999}, [], {"x": 999}, [4.4, 5.5]]
    assert awkward1.to_list(awkward1.fill_none(array, [])) == [[1.1, 2.2, 3.3], [], [], [], [4.4, 5.5]]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/master/LICENSE

from __future__ import absolute_import

import sys

import pytest

import numpy
import awkward1


def test_numpyarray():
    for dtype1 in ("i1", "i2", "i4", "i8", "u1", "u2", "u4", "u8", "f4", "f8", "?"):
        for dtype2 in ("i1", "i2", "i4", "i8", "u1", "u2", "u4", "u8", "f4", "f8", "?"):
            for dtype3 in ("i1", "i2", "i4", "i8", "u1", "u2", "u4", "u8", "f4", "f8", "?"):
                for dtype4 in ("i1", "i2", "i4", "i8", "u1", "u2", "u4", "u8", "f4", "f8", "?"):
                    one   = numpy.array([0, 1, 2], dtype=dtype1)
                    two   = numpy.array([3, 0], dtype=dtype2)
                    three = numpy.array([], dtype=dtype3)
                    four  = numpy.array([4, 5, 0, 6, 7], dtype=dtype4)
                    combined = numpy.concatenate([one, two, three, four])

                    ak_combined = awkward1.layout.NumpyArray(one).mergemany([
                        awkward1.layout.NumpyArray(two),
                        awkward1.layout.NumpyArray(three),
                        awkward1.layout.NumpyArray(four),
                    ])

                    assert awkward1.to_list(ak_combined) == combined.tolist()
                    assert awkward1.to_numpy(ak_combined).dtype == combined.dtype

                    ak_combined = awkward1.layout.NumpyArray(one).mergemany([
                        awkward1.layout.NumpyArray(two),
                        awkward1.layout.EmptyArray(),
                        awkward1.layout.NumpyArray(four),
                    ])

                    assert awkward1.to_list(ak_combined) == combined.tolist()
                    assert awkward1.to_numpy(ak_combined).dtype == numpy.concatenate([one, two, four]).dtype

def test_lists():
    one = awkward1.Array([[1, 2, 3], [], [4, 5]]).layout
    two = awkward1.Array([[1.1, 2.2], [3.3, 4.4]]).layout
    three = awkward1.layout.EmptyArray()
    four = awkward1.from_numpy(numpy.array([[10], [20]]), regulararray=True, highlevel=False)
    assert awkward1.to_list(one.mergemany([two, three, four])) == [[1.0, 2.0, 3.0], [], [4.0, 5.0], [1.1, 2.2], [3.3, 4.4], [10.0], [20.0]]
    assert awkward1.to_list(four.mergemany([three, two, one])) == [[10.0], [20.0], [1.1, 2.2], [3.3, 4.4], [1.0, 2.0, 3.0], [], [4.0, 5.0]]

    one = awkward1.layout.ListArray64(one.starts, one.stops, one.content)
    two = awkward1.layout.ListArray64(two.starts, two.stops, two.content)
    assert awkward1.to_list(one.mergemany([two, three, four])) == [[1.0, 2.0, 3.0], [], [4.0, 5.0], [1.1, 2.2], [3.3, 4.4], [10.0], [20.0]]
    assert awkward1.to_list(four.mergemany([three, two, one])) == [[10.0], [20.0], [1.1, 2.2], [3.3, 4.4], [1.0, 2.0, 3.0], [], [4.0, 5.0]]

def test_records():
    one = awkward1.Array([{"x": 1, "y": [1]}, {"x": 2, "y": [1, 2]}, {"x": 3, "y": [1, 2, 3]}]).layout
    two = awkward1.Array([{"y": [], "x": 4}, {"y": [3, 2, 1], "x": 5}]).layout
    three = two[0:0]
    four = awkward1.Array([{"x": 6, "y": [1]}, {"x": 7, "y": [1, 2]}]).layout
    assert awkward1.to_list(one.mergemany([two, three, four])) == [{"x": 1, "y": [1]}, {"x": 2, "y": [1, 2]}, {"x": 3, "y": [1, 2, 3]}, {"y": [], "x": 4}, {"y": [3, 2, 1], "x": 5}, {"x": 6, "y": [1]}, {"x": 7, "y": [1, 2]}]

    three = awkward1.layout.EmptyArray()
    assert awkward1.to_list(one.mergemany([two, three, four])) == [{"x": 1, "y": [1]}, {"x": 2, "y": [1, 2]}, {"x": 3, "y": [1, 2, 3]}, {"y": [], "x": 4}, {"y": [3, 2, 1], "x": 5}, {"x": 6, "y": [1]}, {"x": 7, "y": [1, 2]}]

def test_tuples():
    one = awkward1.Array([(1, [1]), (2, [1, 2]), (3, [1, 2, 3])]).layout
    two = awkward1.Array([(4, []), (5, [3, 2, 1])]).layout
    three = two[0:0]
    four = awkward1.Array([(6, [1]), (7, [1, 2])]).layout
    assert awkward1.to_list(one.mergemany([two, three, four])) == [(1, [1]), (2, [1, 2]), (3, [1, 2, 3]), (4, []), (5, [3, 2, 1]), (6, [1]), (7, [1, 2])]

    three = awkward1.layout.EmptyArray()
    assert awkward1.to_list(one.mergemany([two, three, four])) == [(1, [1]), (2, [1, 2]), (3, [1, 2, 3]), (4, []), (5, [3, 2, 1]), (6, [1]), (7, [1, 2])]

def test_indexed():
    one = awkward1.Array([1, 2, 3, None, 4, None, None, 5]).layout
    two = awkward1.Array([6, 7, 8]).layout
    three = awkward1.layout.EmptyArray()
    four = awkward1.Array([9, None, None]).layout
    assert awkward1.to_list(one.mergemany([two, three, four])) == [1, 2, 3, None, 4, None, None, 5, 6, 7, 8, 9, None, None]

def test_reverse_indexed():
    one = awkward1.Array([1, 2, 3]).layout
    two = awkward1.Array([4, 5]).layout
    three = awkward1.Array([None, 6, None]).layout
    assert awkward1.to_list(one.mergemany([two, three])) == [1, 2, 3, 4, 5, None, 6, None]

    four = awkward1.Array([7, 8, None, None, 9]).layout
    assert awkward1.to_list(one.mergemany([two, three, four])) == [1, 2, 3, 4, 5, None, 6, None, 7, 8, None, None, 9]

def test_bytemasked():
    one = awkward1.Array([1, 2, 3, 4, 5, 6]).mask[[True, True, False, True, False, True]].layout
    two = awkward1.Array([7, 99, 999, 8, 9]).mask[[True, False, False, True, True]].layout
    three = awkward1.Array([100, 200, 300]).layout
    four = awkward1.Array([None, None, 123, None]).layout
    assert awkward1.to_list(one.mergemany([two, three, four])) == [1, 2, None, 4, None, 6, 7, None, None, 8, 9, 100, 200, 300, None, None, 123, None]
    assert awkward1.to_list(four.mergemany([three, two, one])) == [None, None, 123, None, 100, 200, 300, 7, None, None, 8, 9, 1, 2, None, 4, None, 6]
    assert awkward1.to_list(three.mergemany([four, one])) == [100, 200, 300, None, None, 123, None, 1, 2, None, 4, None, 6]
    assert awkward1.to_list(three.mergemany([four, one, two])) == [100, 200, 300, None, None, 123, None, 1, 2, None, 4, None, 6, 7, None, None, 8, 9]
    assert awkward1.to_list(three.mergemany([two, one])) == [100, 200, 300, 7, None, None, 8, 9, 1, 2, None, 4, None, 6]
    assert awkward1.to_list(three.mergemany([two, one, four])) == [100, 200, 300, 7, None, None, 8, 9, 1, 2, None, 4, None, 6, None, None, 123, None]

def test_empty():
    one = awkward1.layout.EmptyArray()
    two = awkward1.layout.EmptyArray()
    three = awkward1.Array([1, 2, 3]).layout
    four = awkward1.Array([4, 5]).layout
    assert awkward1.to_list(one.mergemany([two])) == []
    assert awkward1.to_list(one.mergemany([two, one, two, one, two])) == []
    assert awkward1.to_list(one.mergemany([two, three])) == [1, 2, 3]
    assert awkward1.to_list(one.mergemany([two, three, four])) == [1, 2, 3, 4, 5]
    assert awkward1.to_list(one.mergemany([three])) == [1, 2, 3]
    assert awkward1.to_list(one.mergemany([three, four])) == [1, 2, 3, 4, 5]
    assert awkward1.to_list(one.mergemany([three, two])) == [1, 2, 3]
    assert awkward1.to_list(one.mergemany([three, two, four])) == [1, 2, 3, 4, 5]
    assert awkward1.to_list(one.mergemany([three, four, two])) == [1, 2, 3, 4, 5]

def test_union():
    one = awkward1.Array([1, 2, [], [3, 4]]).layout
    two = awkward1.Array([100, 200, 300]).layout
    three = awkward1.Array([{"x": 1}, {"x": 2}, 5, 6, 7]).layout

    assert awkward1.to_list(one.mergemany([two, three])) == [1, 2, [], [3, 4], 100, 200, 300, {"x": 1}, {"x": 2}, 5, 6, 7]
    assert awkward1.to_list(one.mergemany([three, two])) == [1, 2, [], [3, 4], {"x": 1}, {"x": 2}, 5, 6, 7, 100, 200, 300]
    assert awkward1.to_list(two.mergemany([one, three])) == [100, 200, 300, 1, 2, [], [3, 4], {"x": 1}, {"x": 2}, 5, 6, 7]
    assert awkward1.to_list(two.mergemany([three, one])) == [100, 200, 300, {"x": 1}, {"x": 2}, 5, 6, 7, 1, 2, [], [3, 4]]
    assert awkward1.to_list(three.mergemany([one, two])) == [{"x": 1}, {"x": 2}, 5, 6, 7, 1, 2, [], [3, 4], 100, 200, 300]
    assert awkward1.to_list(three.mergemany([two, one])) == [{"x": 1}, {"x": 2}, 5, 6, 7, 100, 200, 300, 1, 2, [], [3, 4]]

def test_union_option():
    one = awkward1.Array([1, 2, [], [3, 4]]).layout
    two = awkward1.Array([100, None, 300]).layout
    three = awkward1.Array([{"x": 1}, {"x": 2}, 5, 6, 7]).layout

    assert awkward1.to_list(one.mergemany([two, three])) == [1, 2, [], [3, 4], 100, None, 300, {"x": 1}, {"x": 2}, 5, 6, 7]
    assert awkward1.to_list(one.mergemany([three, two])) == [1, 2, [], [3, 4], {"x": 1}, {"x": 2}, 5, 6, 7, 100, None, 300]
    assert awkward1.to_list(two.mergemany([one, three])) == [100, None, 300, 1, 2, [], [3, 4], {"x": 1}, {"x": 2}, 5, 6, 7]
    assert awkward1.to_list(two.mergemany([three, one])) == [100, None, 300, {"x": 1}, {"x": 2}, 5, 6, 7, 1, 2, [], [3, 4]]
    assert awkward1.to_list(three.mergemany([one, two])) == [{"x": 1}, {"x": 2}, 5, 6, 7, 1, 2, [], [3, 4], 100, None, 300]
    assert awkward1.to_list(three.mergemany([two, one])) == [{"x": 1}, {"x": 2}, 5, 6, 7, 100, None, 300, 1, 2, [], [3, 4]]

    one = awkward1.Array([1, 2, [], [3, 4]]).layout
    two = awkward1.Array([100, None, 300]).layout
    three = awkward1.Array([{"x": 1}, {"x": 2}, 5, None, 7]).layout

    assert awkward1.to_list(one.mergemany([two, three])) == [1, 2, [], [3, 4], 100, None, 300, {"x": 1}, {"x": 2}, 5, None, 7]
    assert awkward1.to_list(one.mergemany([three, two])) == [1, 2, [], [3, 4], {"x": 1}, {"x": 2}, 5, None, 7, 100, None, 300]
    assert awkward1.to_list(two.mergemany([one, three])) == [100, None, 300, 1, 2, [], [3, 4], {"x": 1}, {"x": 2}, 5, None, 7]
    assert awkward1.to_list(two.mergemany([three, one])) == [100, None, 300, {"x": 1}, {"x": 2}, 5, None, 7, 1, 2, [], [3, 4]]
    assert awkward1.to_list(three.mergemany([one, two])) == [{"x": 1}, {"x": 2}, 5, None, 7, 1, 2, [], [3, 4], 100, None, 300]
    assert awkward1.to_list(three.mergemany([two, one])) == [{"x": 1}, {"x": 2}, 5, None, 7, 100, None, 300, 1, 2, [], [3, 4]]

    one = awkward1.Array([1, 2, [], [3, 4]]).layout
    two = awkward1.Array([100, 200, 300]).layout
    three = awkward1.Array([{"x": 1}, {"x": 2}, 5, None, 7]).layout

    assert awkward1.to_list(one.mergemany([two, three])) == [1, 2, [], [3, 4], 100, 200, 300, {"x": 1}, {"x": 2}, 5, None, 7]
    assert awkward1.to_list(one.mergemany([three, two])) == [1, 2, [], [3, 4], {"x": 1}, {"x": 2}, 5, None, 7, 100, 200, 300]
    assert awkward1.to_list(two.mergemany([one, three])) == [100, 200, 300, 1, 2, [], [3, 4], {"x": 1}, {"x": 2}, 5, None, 7]
    assert awkward1.to_list(two.mergemany([three, one])) == [100, 200, 300, {"x": 1}, {"x": 2}, 5, None, 7, 1, 2, [], [3, 4]]
    assert awkward1.to_list(three.mergemany([one, two])) == [{"x": 1}, {"x": 2}, 5, None, 7, 1, 2, [], [3, 4], 100, 200, 300]
    assert awkward1.to_list(three.mergemany([two, one])) == [{"x": 1}, {"x": 2}, 5, None, 7, 100, 200, 300, 1, 2, [], [3, 4]]

def test_strings():
    one = awkward1.Array(["uno", "dos", "tres"]).layout
    two = awkward1.Array(["un", "deux", "trois", "quatre"]).layout
    three = awkward1.Array(["onay", "ootay", "eethray"]).layout
    assert awkward1.to_list(one.mergemany([two, three])) == ["uno", "dos", "tres", "un", "deux", "trois", "quatre", "onay", "ootay", "eethray"]

def test_concatenate():
    one = awkward1.Array([1, 2, 3])
    two = awkward1.Array([4.4, 5.5])
    three = awkward1.Array([6, 7, 8])
    four = awkward1.Array([[9, 9, 9], [10, 10, 10]])
    assert awkward1.concatenate([one, two, three, four]).tolist() == [1, 2, 3, 4.4, 5.5, 6, 7, 8, [9, 9, 9], [10, 10, 10]]
    assert awkward1.concatenate([four, one, two, three]).tolist() == [[9, 9, 9], [10, 10, 10], 1, 2, 3, 4.4, 5.5, 6, 7, 8]
    assert awkward1.concatenate([one, two, four, three]).tolist() == [1, 2, 3, 4.4, 5.5, [9, 9, 9], [10, 10, 10], 6, 7, 8]

    five = awkward1.Array(["nine", "ten"])
    assert awkward1.concatenate([one, two, three, five]).tolist() == [1, 2, 3, 4.4, 5.5, 6, 7, 8, "nine", "ten"]
    assert awkward1.concatenate([five, one, two, three]).tolist() == ["nine", "ten", 1, 2, 3, 4.4, 5.5, 6, 7, 8]
    assert awkward1.concatenate([one, two, five, three]).tolist() == [1, 2, 3, 4.4, 5.5, "nine", "ten", 6, 7, 8]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/master/LICENSE

from __future__ import absolute_import

import sys

import pytest

import awkward1

def test():
    for itype in ["i8", "u8", "i32", "u32", "i64"]:
        form = awkward1.forms.ListOffsetForm(itype, awkward1.forms.EmptyForm())
        assert form.offsets == itype



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/master/LICENSE

from __future__ import absolute_import

import sys

import pytest
import numpy

import awkward1

def test_2d():
    array = awkward1.from_iter([
        [3.3, 2.2, 5.5, 1.1, 4.4],
        [4.4, 2.2, 1.1, 3.3, 5.5],
        [2.2, 1.1, 4.4, 3.3, 5.5]], highlevel=False)
    assert awkward1.to_list(array.argmin(axis=0)) == [2, 2, 1, 0, 0]
    assert awkward1.to_list(array.argmin(axis=1)) == [3, 2, 1]

def test_3d():
    array = awkward1.from_iter([
        [[ 3.3,  2.2,  5.5,  1.1,  4.4],
         [ 4.4,  2.2,  1.1,  3.3,  5.5],
         [ 2.2,  1.1,  4.4,  3.3,  5.5]],
        [[-3.3,  2.2, -5.5,  1.1,  4.4],
         [ 4.4, -2.2,  1.1, -3.3,  5.5],
         [ 2.2,  1.1,  4.4,  3.3, -5.5]]], highlevel=False)
    assert awkward1.to_list(array.argmin(axis=0)) == [
        [1, 0, 1, 0, 0],
        [0, 1, 0, 1, 0],
        [0, 0, 0, 0, 1]]
    assert awkward1.to_list(array.argmin(axis=1)) == [
        [2, 2, 1, 0, 0],
        [0, 1, 0, 1, 2]]
    assert awkward1.to_list(array.argmin(axis=2)) == [
        [3, 2, 1],
        [2, 3, 4]]
    assert awkward1.to_list(array.argmin(axis=-1)) == [
        [3, 2, 1],
        [2, 3, 4]]
    assert awkward1.to_list(array.argmin(axis=-2)) == [
        [2, 2, 1, 0, 0],
        [0, 1, 0, 1, 2]]
    assert awkward1.to_list(array.argmin(axis=-3)) == [
        [1, 0, 1, 0, 0],
        [0, 1, 0, 1, 0],
        [0, 0, 0, 0, 1]]

def test_jagged():
    array = awkward1.from_iter([[2.2, 1.1, 3.3], [], [4.4, 5.5], [5.5], [-4.4, -5.5, -6.6]], highlevel=False)
    assert awkward1.to_list(array.argmin(axis=1)) == [1, None, 0, 0, 2]

    index2 = awkward1.layout.Index64(numpy.array([4, 3, 2, 1, 0], dtype=numpy.int64))
    array2 = awkward1.layout.IndexedArray64(index2, array)
    assert awkward1.to_list(array2.argmin(axis=1)) == [2, 0, 0, None, 1]

    index3 = awkward1.layout.Index64(numpy.array([4, 3, -1, 4, 0], dtype=numpy.int64))
    array2 = awkward1.layout.IndexedArray64(index3, array)
    assert awkward1.to_list(array2.argmin(axis=1)) == [2, 0, None, 2, 1]
    assert awkward1.to_list(array2.argmin(axis=-1)) == [2, 0, None, 2, 1]

def test_missing():
    array = awkward1.from_iter([[[2.2, 1.1, 3.3]], [[]], [None, None, None], [[-4.4, -5.5, -6.6]]], highlevel=False)
    assert awkward1.to_list(array.argmin(axis=2)) == [[1], [None], [None, None, None], [2]]

def test_highlevel():
    array = awkward1.Array([
        [3.3, 1.1,  5.5, 1.1, 4.4],
        [4.4, 2.2,  1.1, 6.6     ],
        [2.2, 3.3, -1.1          ]])
    assert awkward1.argmin(array) == 11
    assert awkward1.argmax(array) == 8
    assert awkward1.to_list(awkward1.argmin(array, axis=0)) == [2, 0, 2, 0, 0]
    assert awkward1.to_list(awkward1.argmax(array, axis=0)) == [1, 2, 0, 1, 0]
    assert awkward1.to_list(awkward1.argmin(array, axis=1)) == [1, 2, 2]
    assert awkward1.to_list(awkward1.argmax(array, axis=1)) == [2, 3, 1]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/master/LICENSE

from __future__ import absolute_import

import sys

import pytest

import numpy
import awkward1


def test_issue434():
    a = awkward1.Array([[0.0, 1.1, 2.2], [3.3, 4.4], [5.5]])
    b = awkward1.Array([[9.9, 8.8, 7.7], [6.6, 5.5], [4.4]])
    assert awkward1.to_list(b[awkward1.argmin(a, axis=1, keepdims=True)]) == [[9.9], [6.6], [4.4]]
    assert awkward1.to_list(b[awkward1.argmax(a, axis=1, keepdims=True)]) == [[7.7], [5.5], [4.4]]


def test_nokeepdims():
    nparray = numpy.arange(2*3*5, dtype=numpy.int64).reshape(2, 3, 5)
    content = awkward1.layout.NumpyArray(numpy.arange(2*3*5, dtype=numpy.int64))
    regular = awkward1.layout.RegularArray(content, 5)
    listoffset = regular.toListOffsetArray64(False)
    regular_regular = awkward1.layout.RegularArray(regular, 3)
    listoffset_regular = regular_regular.toListOffsetArray64(False)
    regular_listoffset = awkward1.layout.RegularArray(listoffset, 3)
    listoffset_listoffset = regular_listoffset.toListOffsetArray64(False)

    assert awkward1.to_list(regular_regular) == [[[0, 1, 2, 3, 4], [5, 6, 7, 8, 9], [10, 11, 12, 13, 14]], [[15, 16, 17, 18, 19], [20, 21, 22, 23, 24], [25, 26, 27, 28, 29]]]
    assert awkward1.to_list(listoffset_regular) == [[[0, 1, 2, 3, 4], [5, 6, 7, 8, 9], [10, 11, 12, 13, 14]], [[15, 16, 17, 18, 19], [20, 21, 22, 23, 24], [25, 26, 27, 28, 29]]]
    assert awkward1.to_list(regular_listoffset) == [[[0, 1, 2, 3, 4], [5, 6, 7, 8, 9], [10, 11, 12, 13, 14]], [[15, 16, 17, 18, 19], [20, 21, 22, 23, 24], [25, 26, 27, 28, 29]]]
    assert awkward1.to_list(listoffset_listoffset) == [[[0, 1, 2, 3, 4], [5, 6, 7, 8, 9], [10, 11, 12, 13, 14]], [[15, 16, 17, 18, 19], [20, 21, 22, 23, 24], [25, 26, 27, 28, 29]]]

    assert str(awkward1.type(awkward1.Array(listoffset_listoffset))) == "2 * var * var * int64"
    axis1 = awkward1.sum(listoffset_listoffset, axis=-1)
    axis2 = awkward1.sum(listoffset_listoffset, axis=-2)
    axis3 = awkward1.sum(listoffset_listoffset, axis=-3)
    assert awkward1.to_list(axis1) == numpy.sum(nparray, axis=-1).tolist()
    assert awkward1.to_list(axis2) == numpy.sum(nparray, axis=-2).tolist()
    assert awkward1.to_list(axis3) == numpy.sum(nparray, axis=-3).tolist()
    assert str(awkward1.type(awkward1.Array(axis1))) == "2 * var * int64"
    assert str(awkward1.type(awkward1.Array(axis2))) == "2 * var * int64"
    assert str(awkward1.type(awkward1.Array(axis3))) == "3 * var * int64"

    assert str(awkward1.type(awkward1.Array(listoffset_regular))) == "2 * var * 5 * int64"
    axis1 = awkward1.sum(listoffset_regular, axis=-1)
    axis2 = awkward1.sum(listoffset_regular, axis=-2)
    axis3 = awkward1.sum(listoffset_regular, axis=-3)
    assert awkward1.to_list(axis1) == numpy.sum(nparray, axis=-1).tolist()
    assert awkward1.to_list(axis2) == numpy.sum(nparray, axis=-2).tolist()
    assert awkward1.to_list(axis3) == numpy.sum(nparray, axis=-3).tolist()
    assert str(awkward1.type(awkward1.Array(axis1))) == "2 * var * int64"
    assert str(awkward1.type(awkward1.Array(axis2))) == "2 * 5 * int64"
    assert str(awkward1.type(awkward1.Array(axis3))) == "3 * 5 * int64"

    assert str(awkward1.type(awkward1.Array(regular_listoffset))) == "2 * 3 * var * int64"
    axis1 = awkward1.sum(regular_listoffset, axis=-1)
    axis2 = awkward1.sum(regular_listoffset, axis=-2)
    axis3 = awkward1.sum(regular_listoffset, axis=-3)
    assert awkward1.to_list(axis1) == numpy.sum(nparray, axis=-1).tolist()
    assert awkward1.to_list(axis2) == numpy.sum(nparray, axis=-2).tolist()
    assert awkward1.to_list(axis3) == numpy.sum(nparray, axis=-3).tolist()
    assert str(awkward1.type(awkward1.Array(axis1))) == "2 * 3 * int64"
    assert str(awkward1.type(awkward1.Array(axis2))) == "2 * var * int64"
    assert str(awkward1.type(awkward1.Array(axis3))) == "3 * var * int64"

    assert str(awkward1.type(awkward1.Array(regular_regular))) == "2 * 3 * 5 * int64"
    axis1 = awkward1.sum(regular_regular, axis=-1)
    axis2 = awkward1.sum(regular_regular, axis=-2)
    axis3 = awkward1.sum(regular_regular, axis=-3)
    assert awkward1.to_list(axis1) == numpy.sum(nparray, axis=-1).tolist()
    assert awkward1.to_list(axis2) == numpy.sum(nparray, axis=-2).tolist()
    assert awkward1.to_list(axis3) == numpy.sum(nparray, axis=-3).tolist()
    assert str(awkward1.type(awkward1.Array(axis1))) == "2 * 3 * int64"
    assert str(awkward1.type(awkward1.Array(axis2))) == "2 * 5 * int64"
    assert str(awkward1.type(awkward1.Array(axis3))) == "3 * 5 * int64"


def test_keepdims():
    nparray = numpy.arange(2*3*5, dtype=numpy.int64).reshape(2, 3, 5)
    content = awkward1.layout.NumpyArray(numpy.arange(2*3*5, dtype=numpy.int64))
    regular = awkward1.layout.RegularArray(content, 5)
    listoffset = regular.toListOffsetArray64(False)
    regular_regular = awkward1.layout.RegularArray(regular, 3)
    listoffset_regular = regular_regular.toListOffsetArray64(False)
    regular_listoffset = awkward1.layout.RegularArray(listoffset, 3)
    listoffset_listoffset = regular_listoffset.toListOffsetArray64(False)

    assert str(awkward1.type(awkward1.Array(listoffset_listoffset))) == "2 * var * var * int64"
    axis1 = awkward1.sum(listoffset_listoffset, axis=-1, keepdims=True)
    axis2 = awkward1.sum(listoffset_listoffset, axis=-2, keepdims=True)
    axis3 = awkward1.sum(listoffset_listoffset, axis=-3, keepdims=True)
    assert awkward1.to_list(axis1) == numpy.sum(nparray, axis=-1, keepdims=True).tolist()
    assert awkward1.to_list(axis2) == numpy.sum(nparray, axis=-2, keepdims=True).tolist()
    assert awkward1.to_list(axis3) == numpy.sum(nparray, axis=-3, keepdims=True).tolist()
    assert str(awkward1.type(awkward1.Array(axis1))) == "2 * var * var * int64"
    assert str(awkward1.type(awkward1.Array(axis2))) == "2 * var * var * int64"
    assert str(awkward1.type(awkward1.Array(axis3))) == "1 * var * var * int64"

    assert str(awkward1.type(awkward1.Array(listoffset_regular))) == "2 * var * 5 * int64"
    axis1 = awkward1.sum(listoffset_regular, axis=-1, keepdims=True)
    axis2 = awkward1.sum(listoffset_regular, axis=-2, keepdims=True)
    axis3 = awkward1.sum(listoffset_regular, axis=-3, keepdims=True)
    assert awkward1.to_list(axis1) == numpy.sum(nparray, axis=-1, keepdims=True).tolist()
    assert awkward1.to_list(axis2) == numpy.sum(nparray, axis=-2, keepdims=True).tolist()
    assert awkward1.to_list(axis3) == numpy.sum(nparray, axis=-3, keepdims=True).tolist()
    assert str(awkward1.type(awkward1.Array(axis1))) == "2 * var * 1 * int64"
    assert str(awkward1.type(awkward1.Array(axis2))) == "2 * var * 5 * int64"
    assert str(awkward1.type(awkward1.Array(axis3))) == "1 * var * 5 * int64"

    assert str(awkward1.type(awkward1.Array(regular_listoffset))) == "2 * 3 * var * int64"
    axis1 = awkward1.sum(regular_listoffset, axis=-1, keepdims=True)
    axis2 = awkward1.sum(regular_listoffset, axis=-2, keepdims=True)
    axis3 = awkward1.sum(regular_listoffset, axis=-3, keepdims=True)
    assert awkward1.to_list(axis1) == numpy.sum(nparray, axis=-1, keepdims=True).tolist()
    assert awkward1.to_list(axis2) == numpy.sum(nparray, axis=-2, keepdims=True).tolist()
    assert awkward1.to_list(axis3) == numpy.sum(nparray, axis=-3, keepdims=True).tolist()
    assert str(awkward1.type(awkward1.Array(axis1))) == "2 * 3 * var * int64"
    assert str(awkward1.type(awkward1.Array(axis2))) == "2 * 1 * var * int64"
    assert str(awkward1.type(awkward1.Array(axis3))) == "1 * 3 * var * int64"

    assert str(awkward1.type(awkward1.Array(regular_regular))) == "2 * 3 * 5 * int64"
    axis1 = awkward1.sum(regular_regular, axis=-1, keepdims=True)
    axis2 = awkward1.sum(regular_regular, axis=-2, keepdims=True)
    axis3 = awkward1.sum(regular_regular, axis=-3, keepdims=True)
    assert awkward1.to_list(axis1) == numpy.sum(nparray, axis=-1, keepdims=True).tolist()
    assert awkward1.to_list(axis2) == numpy.sum(nparray, axis=-2, keepdims=True).tolist()
    assert awkward1.to_list(axis3) == numpy.sum(nparray, axis=-3, keepdims=True).tolist()
    assert str(awkward1.type(awkward1.Array(axis1))) == "2 * 3 * 1 * int64"
    assert str(awkward1.type(awkward1.Array(axis2))) == "2 * 1 * 5 * int64"
    assert str(awkward1.type(awkward1.Array(axis3))) == "1 * 3 * 5 * int64"


def test_nokeepdims_none1():
    content = awkward1.Array([0, 1, 2, None, 4, 5, None, None, 8, 9, 10, 11, 12, None, 14, 15, 16, 17, 18, None, None, None, None, None, None, 25, 26, 27, 28, 29]).layout
    regular = awkward1.layout.RegularArray(content, 5)
    listoffset = regular.toListOffsetArray64(False)
    regular_regular = awkward1.layout.RegularArray(regular, 3)
    listoffset_regular = regular_regular.toListOffsetArray64(False)
    regular_listoffset = awkward1.layout.RegularArray(listoffset, 3)
    listoffset_listoffset = regular_listoffset.toListOffsetArray64(False)

    assert str(awkward1.type(awkward1.Array(listoffset_listoffset))) == "2 * var * var * ?int64"
    axis1 = awkward1.sum(listoffset_listoffset, axis=-1)
    axis2 = awkward1.sum(listoffset_listoffset, axis=-2)
    axis3 = awkward1.sum(listoffset_listoffset, axis=-3)
    assert str(awkward1.type(awkward1.Array(axis1))) == "2 * var * int64"
    assert str(awkward1.type(awkward1.Array(axis2))) == "2 * var * int64"
    assert str(awkward1.type(awkward1.Array(axis3))) == "3 * var * int64"

    assert str(awkward1.type(awkward1.Array(listoffset_regular))) == "2 * var * 5 * ?int64"
    axis1 = awkward1.sum(listoffset_regular, axis=-1)
    axis2 = awkward1.sum(listoffset_regular, axis=-2)
    axis3 = awkward1.sum(listoffset_regular, axis=-3)
    assert str(awkward1.type(awkward1.Array(axis1))) == "2 * var * int64"
    assert str(awkward1.type(awkward1.Array(axis2))) == "2 * var * int64"
    assert str(awkward1.type(awkward1.Array(axis3))) == "3 * var * int64"

    assert str(awkward1.type(awkward1.Array(regular_listoffset))) == "2 * 3 * var * ?int64"
    axis1 = awkward1.sum(regular_listoffset, axis=-1)
    axis2 = awkward1.sum(regular_listoffset, axis=-2)
    axis3 = awkward1.sum(regular_listoffset, axis=-3)
    assert str(awkward1.type(awkward1.Array(axis1))) == "2 * 3 * int64"
    assert str(awkward1.type(awkward1.Array(axis2))) == "2 * var * int64"
    assert str(awkward1.type(awkward1.Array(axis3))) == "3 * var * int64"

    assert str(awkward1.type(awkward1.Array(regular_regular))) == "2 * 3 * 5 * ?int64"
    axis1 = awkward1.sum(regular_regular, axis=-1)
    axis2 = awkward1.sum(regular_regular, axis=-2)
    axis3 = awkward1.sum(regular_regular, axis=-3)
    assert str(awkward1.type(awkward1.Array(axis1))) == "2 * 3 * int64"
    assert str(awkward1.type(awkward1.Array(axis2))) == "2 * var * int64"
    assert str(awkward1.type(awkward1.Array(axis3))) == "3 * var * int64"


def test_keepdims_none1():
    content = awkward1.Array([0, 1, 2, None, 4, 5, None, None, 8, 9, 10, 11, 12, None, 14, 15, 16, 17, 18, None, None, None, None, None, None, 25, 26, 27, 28, 29]).layout
    regular = awkward1.layout.RegularArray(content, 5)
    listoffset = regular.toListOffsetArray64(False)
    regular_regular = awkward1.layout.RegularArray(regular, 3)
    listoffset_regular = regular_regular.toListOffsetArray64(False)
    regular_listoffset = awkward1.layout.RegularArray(listoffset, 3)
    listoffset_listoffset = regular_listoffset.toListOffsetArray64(False)

    assert str(awkward1.type(awkward1.Array(listoffset_listoffset))) == "2 * var * var * ?int64"
    axis1 = awkward1.sum(listoffset_listoffset, axis=-1, keepdims=True)
    axis2 = awkward1.sum(listoffset_listoffset, axis=-2, keepdims=True)
    axis3 = awkward1.sum(listoffset_listoffset, axis=-3, keepdims=True)
    assert str(awkward1.type(awkward1.Array(axis1))) == "2 * var * var * int64"
    assert str(awkward1.type(awkward1.Array(axis2))) == "2 * var * var * int64"
    assert str(awkward1.type(awkward1.Array(axis3))) == "1 * var * var * int64"

    assert str(awkward1.type(awkward1.Array(listoffset_regular))) == "2 * var * 5 * ?int64"
    axis1 = awkward1.sum(listoffset_regular, axis=-1, keepdims=True)
    axis2 = awkward1.sum(listoffset_regular, axis=-2, keepdims=True)
    axis3 = awkward1.sum(listoffset_regular, axis=-3, keepdims=True)
    assert str(awkward1.type(awkward1.Array(axis1))) == "2 * var * var * int64"
    assert str(awkward1.type(awkward1.Array(axis2))) == "2 * var * var * int64"
    assert str(awkward1.type(awkward1.Array(axis3))) == "1 * var * var * int64"

    assert str(awkward1.type(awkward1.Array(regular_listoffset))) == "2 * 3 * var * ?int64"
    axis1 = awkward1.sum(regular_listoffset, axis=-1, keepdims=True)
    axis2 = awkward1.sum(regular_listoffset, axis=-2, keepdims=True)
    axis3 = awkward1.sum(regular_listoffset, axis=-3, keepdims=True)
    assert str(awkward1.type(awkward1.Array(axis1))) == "2 * 3 * var * int64"
    assert str(awkward1.type(awkward1.Array(axis2))) == "2 * 1 * var * int64"
    assert str(awkward1.type(awkward1.Array(axis3))) == "1 * 3 * var * int64"

    assert str(awkward1.type(awkward1.Array(regular_regular))) == "2 * 3 * 5 * ?int64"
    axis1 = awkward1.sum(regular_regular, axis=-1, keepdims=True)
    axis2 = awkward1.sum(regular_regular, axis=-2, keepdims=True)
    axis3 = awkward1.sum(regular_regular, axis=-3, keepdims=True)
    assert str(awkward1.type(awkward1.Array(axis1))) == "2 * 3 * var * int64"
    assert str(awkward1.type(awkward1.Array(axis2))) == "2 * 1 * var * int64"
    assert str(awkward1.type(awkward1.Array(axis3))) == "1 * 3 * var * int64"


def test_nokeepdims_mask1():
    mask = awkward1.layout.Index8(numpy.array([False, False, False, True, False, False, True, True, False, False, False, False, False, True, False, False, False, False, False, True, True, True, True, True, True, False, False, False, False, False]))
    content = awkward1.layout.ByteMaskedArray(mask, awkward1.layout.NumpyArray(numpy.arange(2*3*5, dtype=numpy.int64)), valid_when=False)
    regular = awkward1.layout.RegularArray(content, 5)
    listoffset = regular.toListOffsetArray64(False)
    regular_regular = awkward1.layout.RegularArray(regular, 3)
    listoffset_regular = regular_regular.toListOffsetArray64(False)
    regular_listoffset = awkward1.layout.RegularArray(listoffset, 3)
    listoffset_listoffset = regular_listoffset.toListOffsetArray64(False)

    assert str(awkward1.type(awkward1.Array(listoffset_listoffset))) == "2 * var * var * ?int64"
    axis1 = awkward1.sum(listoffset_listoffset, axis=-1)
    axis2 = awkward1.sum(listoffset_listoffset, axis=-2)
    axis3 = awkward1.sum(listoffset_listoffset, axis=-3)
    assert str(awkward1.type(awkward1.Array(axis1))) == "2 * var * int64"
    assert str(awkward1.type(awkward1.Array(axis2))) == "2 * var * int64"
    assert str(awkward1.type(awkward1.Array(axis3))) == "3 * var * int64"

    assert str(awkward1.type(awkward1.Array(listoffset_regular))) == "2 * var * 5 * ?int64"
    axis1 = awkward1.sum(listoffset_regular, axis=-1)
    axis2 = awkward1.sum(listoffset_regular, axis=-2)
    axis3 = awkward1.sum(listoffset_regular, axis=-3)
    assert str(awkward1.type(awkward1.Array(axis1))) == "2 * var * int64"
    assert str(awkward1.type(awkward1.Array(axis2))) == "2 * var * int64"
    assert str(awkward1.type(awkward1.Array(axis3))) == "3 * var * int64"

    assert str(awkward1.type(awkward1.Array(regular_listoffset))) == "2 * 3 * var * ?int64"
    axis1 = awkward1.sum(regular_listoffset, axis=-1)
    axis2 = awkward1.sum(regular_listoffset, axis=-2)
    axis3 = awkward1.sum(regular_listoffset, axis=-3)
    assert str(awkward1.type(awkward1.Array(axis1))) == "2 * 3 * int64"
    assert str(awkward1.type(awkward1.Array(axis2))) == "2 * var * int64"
    assert str(awkward1.type(awkward1.Array(axis3))) == "3 * var * int64"

    assert str(awkward1.type(awkward1.Array(regular_regular))) == "2 * 3 * 5 * ?int64"
    axis1 = awkward1.sum(regular_regular, axis=-1)
    axis2 = awkward1.sum(regular_regular, axis=-2)
    axis3 = awkward1.sum(regular_regular, axis=-3)
    assert str(awkward1.type(awkward1.Array(axis1))) == "2 * 3 * int64"
    assert str(awkward1.type(awkward1.Array(axis2))) == "2 * var * int64"
    assert str(awkward1.type(awkward1.Array(axis3))) == "3 * var * int64"


def test_keepdims_mask1():
    mask = awkward1.layout.Index8(numpy.array([False, False, False, True, False, False, True, True, False, False, False, False, False, True, False, False, False, False, False, True, True, True, True, True, True, False, False, False, False, False]))
    content = awkward1.layout.ByteMaskedArray(mask, awkward1.layout.NumpyArray(numpy.arange(2*3*5, dtype=numpy.int64)), valid_when=False)
    regular = awkward1.layout.RegularArray(content, 5)
    listoffset = regular.toListOffsetArray64(False)
    regular_regular = awkward1.layout.RegularArray(regular, 3)
    listoffset_regular = regular_regular.toListOffsetArray64(False)
    regular_listoffset = awkward1.layout.RegularArray(listoffset, 3)
    listoffset_listoffset = regular_listoffset.toListOffsetArray64(False)

    assert str(awkward1.type(awkward1.Array(listoffset_listoffset))) == "2 * var * var * ?int64"
    axis1 = awkward1.sum(listoffset_listoffset, axis=-1, keepdims=True)
    axis2 = awkward1.sum(listoffset_listoffset, axis=-2, keepdims=True)
    axis3 = awkward1.sum(listoffset_listoffset, axis=-3, keepdims=True)
    assert str(awkward1.type(awkward1.Array(axis1))) == "2 * var * var * int64"
    assert str(awkward1.type(awkward1.Array(axis2))) == "2 * var * var * int64"
    assert str(awkward1.type(awkward1.Array(axis3))) == "1 * var * var * int64"

    assert str(awkward1.type(awkward1.Array(listoffset_regular))) == "2 * var * 5 * ?int64"
    axis1 = awkward1.sum(listoffset_regular, axis=-1, keepdims=True)
    axis2 = awkward1.sum(listoffset_regular, axis=-2, keepdims=True)
    axis3 = awkward1.sum(listoffset_regular, axis=-3, keepdims=True)
    assert str(awkward1.type(awkward1.Array(axis1))) == "2 * var * var * int64"
    assert str(awkward1.type(awkward1.Array(axis2))) == "2 * var * var * int64"
    assert str(awkward1.type(awkward1.Array(axis3))) == "1 * var * var * int64"

    assert str(awkward1.type(awkward1.Array(regular_listoffset))) == "2 * 3 * var * ?int64"
    axis1 = awkward1.sum(regular_listoffset, axis=-1, keepdims=True)
    axis2 = awkward1.sum(regular_listoffset, axis=-2, keepdims=True)
    axis3 = awkward1.sum(regular_listoffset, axis=-3, keepdims=True)
    assert str(awkward1.type(awkward1.Array(axis1))) == "2 * 3 * var * int64"
    assert str(awkward1.type(awkward1.Array(axis2))) == "2 * 1 * var * int64"
    assert str(awkward1.type(awkward1.Array(axis3))) == "1 * 3 * var * int64"

    assert str(awkward1.type(awkward1.Array(regular_regular))) == "2 * 3 * 5 * ?int64"
    axis1 = awkward1.sum(regular_regular, axis=-1, keepdims=True)
    axis2 = awkward1.sum(regular_regular, axis=-2, keepdims=True)
    axis3 = awkward1.sum(regular_regular, axis=-3, keepdims=True)
    assert str(awkward1.type(awkward1.Array(axis1))) == "2 * 3 * var * int64"
    assert str(awkward1.type(awkward1.Array(axis2))) == "2 * 1 * var * int64"
    assert str(awkward1.type(awkward1.Array(axis3))) == "1 * 3 * var * int64"


def test_nokeepdims_mask2():
    content = awkward1.layout.NumpyArray(numpy.arange(2*3*5, dtype=numpy.int64))
    regular = awkward1.layout.RegularArray(content, 5)
    listoffset = regular.toListOffsetArray64(False)
    mask = awkward1.layout.Index8(numpy.array([False, False, True, True, False, True]))
    regular_regular = awkward1.layout.RegularArray(awkward1.layout.ByteMaskedArray(mask, regular, valid_when=False), 3)
    listoffset_regular = regular_regular.toListOffsetArray64(False)
    regular_listoffset = awkward1.layout.RegularArray(awkward1.layout.ByteMaskedArray(mask, listoffset, valid_when=False), 3)
    listoffset_listoffset = regular_listoffset.toListOffsetArray64(False)

    assert str(awkward1.type(awkward1.Array(listoffset_listoffset))) == "2 * var * option[var * int64]"
    axis1 = awkward1.sum(listoffset_listoffset, axis=-1)
    axis2 = awkward1.sum(listoffset_listoffset, axis=-2)
    axis3 = awkward1.sum(listoffset_listoffset, axis=-3)
    assert str(awkward1.type(awkward1.Array(axis1))) == "2 * var * ?int64"
    assert str(awkward1.type(awkward1.Array(axis2))) == "2 * var * int64"
    assert str(awkward1.type(awkward1.Array(axis3))) == "3 * var * int64"

    assert str(awkward1.type(awkward1.Array(listoffset_regular))) == "2 * var * option[5 * int64]"
    axis1 = awkward1.sum(listoffset_regular, axis=-1)
    axis2 = awkward1.sum(listoffset_regular, axis=-2)
    axis3 = awkward1.sum(listoffset_regular, axis=-3)
    assert str(awkward1.type(awkward1.Array(axis1))) == "2 * var * ?int64"
    assert str(awkward1.type(awkward1.Array(axis2))) == "2 * var * int64"
    assert str(awkward1.type(awkward1.Array(axis3))) == "3 * var * int64"

    assert str(awkward1.type(awkward1.Array(regular_listoffset))) == "2 * 3 * option[var * int64]"
    axis1 = awkward1.sum(regular_listoffset, axis=-1)
    axis2 = awkward1.sum(regular_listoffset, axis=-2)
    axis3 = awkward1.sum(regular_listoffset, axis=-3)
    assert str(awkward1.type(awkward1.Array(axis1))) == "2 * var * ?int64"
    assert str(awkward1.type(awkward1.Array(axis2))) == "2 * var * int64"
    assert str(awkward1.type(awkward1.Array(axis3))) == "3 * var * int64"

    assert str(awkward1.type(awkward1.Array(regular_regular))) == "2 * 3 * option[5 * int64]"
    axis1 = awkward1.sum(regular_regular, axis=-1)
    axis2 = awkward1.sum(regular_regular, axis=-2)
    axis3 = awkward1.sum(regular_regular, axis=-3)
    assert str(awkward1.type(awkward1.Array(axis1))) == "2 * var * ?int64"
    assert str(awkward1.type(awkward1.Array(axis2))) == "2 * var * int64"
    assert str(awkward1.type(awkward1.Array(axis3))) == "3 * var * int64"


def test_keepdims_mask2():
    content = awkward1.layout.NumpyArray(numpy.arange(2*3*5, dtype=numpy.int64))
    regular = awkward1.layout.RegularArray(content, 5)
    listoffset = regular.toListOffsetArray64(False)
    mask = awkward1.layout.Index8(numpy.array([False, False, True, True, False, True]))
    regular_regular = awkward1.layout.RegularArray(awkward1.layout.ByteMaskedArray(mask, regular, valid_when=False), 3)
    listoffset_regular = regular_regular.toListOffsetArray64(False)
    regular_listoffset = awkward1.layout.RegularArray(awkward1.layout.ByteMaskedArray(mask, listoffset, valid_when=False), 3)
    listoffset_listoffset = regular_listoffset.toListOffsetArray64(False)

    assert str(awkward1.type(awkward1.Array(listoffset_listoffset))) == "2 * var * option[var * int64]"
    axis1 = awkward1.sum(listoffset_listoffset, axis=-1, keepdims=True)
    axis2 = awkward1.sum(listoffset_listoffset, axis=-2, keepdims=True)
    axis3 = awkward1.sum(listoffset_listoffset, axis=-3, keepdims=True)
    assert str(awkward1.type(awkward1.Array(axis1))) == "2 * var * option[var * int64]"
    assert str(awkward1.type(awkward1.Array(axis2))) == "2 * var * var * int64"
    assert str(awkward1.type(awkward1.Array(axis3))) == "1 * var * var * int64"

    assert str(awkward1.type(awkward1.Array(listoffset_regular))) == "2 * var * option[5 * int64]"
    axis1 = awkward1.sum(listoffset_regular, axis=-1, keepdims=True)
    axis2 = awkward1.sum(listoffset_regular, axis=-2, keepdims=True)
    axis3 = awkward1.sum(listoffset_regular, axis=-3, keepdims=True)
    assert str(awkward1.type(awkward1.Array(axis1))) == "2 * var * option[1 * int64]"
    assert str(awkward1.type(awkward1.Array(axis2))) == "2 * var * var * int64"
    assert str(awkward1.type(awkward1.Array(axis3))) == "1 * var * var * int64"

    assert str(awkward1.type(awkward1.Array(regular_listoffset))) == "2 * 3 * option[var * int64]"
    axis1 = awkward1.sum(regular_listoffset, axis=-1, keepdims=True)
    axis2 = awkward1.sum(regular_listoffset, axis=-2, keepdims=True)
    axis3 = awkward1.sum(regular_listoffset, axis=-3, keepdims=True)
    assert str(awkward1.type(awkward1.Array(axis1))) == "2 * var * option[var * int64]"
    assert str(awkward1.type(awkward1.Array(axis2))) == "2 * var * var * int64"
    assert str(awkward1.type(awkward1.Array(axis3))) == "1 * var * var * int64"

    assert str(awkward1.type(awkward1.Array(regular_regular))) == "2 * 3 * option[5 * int64]"
    axis1 = awkward1.sum(regular_regular, axis=-1, keepdims=True)
    axis2 = awkward1.sum(regular_regular, axis=-2, keepdims=True)
    axis3 = awkward1.sum(regular_regular, axis=-3, keepdims=True)
    assert str(awkward1.type(awkward1.Array(axis1))) == "2 * var * option[1 * int64]"
    assert str(awkward1.type(awkward1.Array(axis2))) == "2 * var * var * int64"
    assert str(awkward1.type(awkward1.Array(axis3))) == "1 * var * var * int64"


def test_nokeepdims_mask3():
    content = awkward1.layout.NumpyArray(numpy.arange(2*3*5, dtype=numpy.int64))
    regular = awkward1.layout.RegularArray(content, 5)
    listoffset = regular.toListOffsetArray64(False)
    regular_regular = awkward1.layout.RegularArray(regular, 3)
    listoffset_regular = regular_regular.toListOffsetArray64(False)
    regular_listoffset = awkward1.layout.RegularArray(listoffset, 3)
    listoffset_listoffset = regular_listoffset.toListOffsetArray64(False)
    mask = awkward1.layout.Index8(numpy.array([True, False]))
    regular_regular = awkward1.layout.ByteMaskedArray(mask, regular_regular, valid_when=False)
    listoffset_regular = awkward1.layout.ByteMaskedArray(mask, listoffset_regular, valid_when=False)
    regular_listoffset = awkward1.layout.ByteMaskedArray(mask, regular_listoffset, valid_when=False)
    listoffset_listoffset = awkward1.layout.ByteMaskedArray(mask, listoffset_listoffset, valid_when=False)

    assert str(awkward1.type(awkward1.Array(listoffset_listoffset))) == "2 * option[var * var * int64]"
    axis1 = awkward1.sum(listoffset_listoffset, axis=-1)
    axis2 = awkward1.sum(listoffset_listoffset, axis=-2)
    axis3 = awkward1.sum(listoffset_listoffset, axis=-3)
    assert str(awkward1.type(awkward1.Array(axis1))) == "2 * option[var * int64]"
    assert str(awkward1.type(awkward1.Array(axis2))) == "2 * option[var * int64]"
    assert str(awkward1.type(awkward1.Array(axis3))) == "3 * var * int64"

    assert str(awkward1.type(awkward1.Array(listoffset_regular))) == "2 * option[var * 5 * int64]"
    axis1 = awkward1.sum(listoffset_regular, axis=-1)
    axis2 = awkward1.sum(listoffset_regular, axis=-2)
    axis3 = awkward1.sum(listoffset_regular, axis=-3)
    assert str(awkward1.type(awkward1.Array(axis1))) == "2 * option[var * int64]"
    assert str(awkward1.type(awkward1.Array(axis2))) == "2 * option[5 * int64]"
    assert str(awkward1.type(awkward1.Array(axis3))) == "3 * 5 * int64"

    assert str(awkward1.type(awkward1.Array(regular_listoffset))) == "2 * option[3 * var * int64]"
    axis1 = awkward1.sum(regular_listoffset, axis=-1)
    axis2 = awkward1.sum(regular_listoffset, axis=-2)
    axis3 = awkward1.sum(regular_listoffset, axis=-3)
    assert str(awkward1.type(awkward1.Array(axis1))) == "2 * option[var * int64]"
    assert str(awkward1.type(awkward1.Array(axis2))) == "2 * option[var * int64]"
    assert str(awkward1.type(awkward1.Array(axis3))) == "3 * var * int64"

    assert str(awkward1.type(awkward1.Array(regular_regular))) == "2 * option[3 * 5 * int64]"
    axis1 = awkward1.sum(regular_regular, axis=-1)
    axis2 = awkward1.sum(regular_regular, axis=-2)
    axis3 = awkward1.sum(regular_regular, axis=-3)
    assert str(awkward1.type(awkward1.Array(axis1))) == "2 * option[var * int64]"
    assert str(awkward1.type(awkward1.Array(axis2))) == "2 * option[5 * int64]"
    assert str(awkward1.type(awkward1.Array(axis3))) == "3 * 5 * int64"


def test_keepdims_mask3():
    content = awkward1.layout.NumpyArray(numpy.arange(2*3*5, dtype=numpy.int64))
    regular = awkward1.layout.RegularArray(content, 5)
    listoffset = regular.toListOffsetArray64(False)
    regular_regular = awkward1.layout.RegularArray(regular, 3)
    listoffset_regular = regular_regular.toListOffsetArray64(False)
    regular_listoffset = awkward1.layout.RegularArray(listoffset, 3)
    listoffset_listoffset = regular_listoffset.toListOffsetArray64(False)
    mask = awkward1.layout.Index8(numpy.array([True, False]))
    regular_regular = awkward1.layout.ByteMaskedArray(mask, regular_regular, valid_when=False)
    listoffset_regular = awkward1.layout.ByteMaskedArray(mask, listoffset_regular, valid_when=False)
    regular_listoffset = awkward1.layout.ByteMaskedArray(mask, regular_listoffset, valid_when=False)
    listoffset_listoffset = awkward1.layout.ByteMaskedArray(mask, listoffset_listoffset, valid_when=False)

    assert str(awkward1.type(awkward1.Array(listoffset_listoffset))) == "2 * option[var * var * int64]"
    axis1 = awkward1.sum(listoffset_listoffset, axis=-1, keepdims=True)
    axis2 = awkward1.sum(listoffset_listoffset, axis=-2, keepdims=True)
    axis3 = awkward1.sum(listoffset_listoffset, axis=-3, keepdims=True)
    assert str(awkward1.type(awkward1.Array(axis1))) == "2 * option[var * var * int64]"
    assert str(awkward1.type(awkward1.Array(axis2))) == "2 * option[var * var * int64]"
    assert str(awkward1.type(awkward1.Array(axis3))) == "1 * var * var * int64"

    assert str(awkward1.type(awkward1.Array(listoffset_regular))) == "2 * option[var * 5 * int64]"
    axis1 = awkward1.sum(listoffset_regular, axis=-1, keepdims=True)
    axis2 = awkward1.sum(listoffset_regular, axis=-2, keepdims=True)
    axis3 = awkward1.sum(listoffset_regular, axis=-3, keepdims=True)
    assert str(awkward1.type(awkward1.Array(axis1))) == "2 * option[var * 1 * int64]"
    assert str(awkward1.type(awkward1.Array(axis2))) == "2 * option[var * 5 * int64]"
    assert str(awkward1.type(awkward1.Array(axis3))) == "1 * var * 5 * int64"

    assert str(awkward1.type(awkward1.Array(regular_listoffset))) == "2 * option[3 * var * int64]"
    axis1 = awkward1.sum(regular_listoffset, axis=-1, keepdims=True)
    axis2 = awkward1.sum(regular_listoffset, axis=-2, keepdims=True)
    axis3 = awkward1.sum(regular_listoffset, axis=-3, keepdims=True)
    assert str(awkward1.type(awkward1.Array(axis1))) == "2 * option[var * var * int64]"
    assert str(awkward1.type(awkward1.Array(axis2))) == "2 * option[var * var * int64]"
    assert str(awkward1.type(awkward1.Array(axis3))) == "1 * var * var * int64"

    assert str(awkward1.type(awkward1.Array(regular_regular))) == "2 * option[3 * 5 * int64]"
    axis1 = awkward1.sum(regular_regular, axis=-1, keepdims=True)
    axis2 = awkward1.sum(regular_regular, axis=-2, keepdims=True)
    axis3 = awkward1.sum(regular_regular, axis=-3, keepdims=True)
    assert str(awkward1.type(awkward1.Array(axis1))) == "2 * option[var * 1 * int64]"
    assert str(awkward1.type(awkward1.Array(axis2))) == "2 * option[var * 5 * int64]"
    assert str(awkward1.type(awkward1.Array(axis3))) == "1 * var * 5 * int64"
